# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from pyxem.detectors.generic_flat_detector import GenericFlatDetector
from pyFAI.detectors import Detector


def test_generic_flat_detector_init():
    size_x, size_y = 256, 256
    detector = GenericFlatDetector(size_x, size_y)
    assert isinstance(detector, Detector)
    assert np.array_equal(detector.max_shape, (size_x, size_y))
