name = "pyxem"
version = "0.11.0"
author = "Duncan Johnstone, Phillip Crout"
copyright = "copyright 2016-2020, the pyxem developers"
credits = [
    "Duncan Johnstone",
    "Phillip Crout",
    "Joonatan Laulainen",
    "Simon Høgås",
    "Ben Martineau",
    "Tina Bergh",
    "Stef Smeets",
    "Carter Francis",
    "Eirik Opheim",
    "Eric Prestat",
    "Sean Collins",
    "Tom Furnival",
    "Mohsen Danaie",
    "Håkon Wiik Ånes",
    "Tiarnan Doherty",
    "Affan Iqbal",
    "Jedrzej Morzy",
    "Tomas Ostasevicius",
    "Rob Tovey",
    "Matt von Lany",
    "Tim Poon",
    "Endre Jacobsen",
]
license = "GPLv3"
maintainer = "Duncan Johnstone, Phillip Crout"
email = "pyxem.team@gmail.com"
status = "development"
